from abc import ABC

from beanie.odm.operators.find import BaseFindOperator


class BaseFindArrayOperator(BaseFindOperator, ABC):
    ...


class All(BaseFindArrayOperator):
    """
    `$all` array query operator

    Example:

    ```python
    class Sample(Document):
        results: List[int]

    All(Sample.results, [80, 85])
    ```

    Will return query object like

    ```python
    {"results": {"$all": [80, 85]}}
    ```

    MongoDB doc:
    <https://docs.mongodb.com/manual/reference/operator/query/all>
    """

    def __init__(
        self,
        field,
        values: list,
    ):
        self.field = field
        self.values_list = values

    @property
    def query(self):
        return {self.field: {"$all": self.values_list}}


class ElemMatch(BaseFindArrayOperator):
    """
    `$elemMatch` array query operator

    Example:

    ```python
    class Sample(Document):
        results: List[int]

    ElemMatch(Sample.results, [80, 85])
    ```

    Will return query object like

    ```python
    {"results": {"$elemMatch": [80, 85]}}
    ```

    MongoDB doc:
    <https://docs.mongodb.com/manual/reference/operator/query/elemMatch/>
    """

    def __init__(
        self,
        field,
        expression: dict,
    ):
        self.field = field
        self.expression = expression

    @property
    def query(self):
        return {self.field: {"$elemMatch": self.expression}}


class Size(BaseFindArrayOperator):
    """
    `$size` array query operator

    Example:

    ```python
    class Sample(Document):
        results: List[int]

    Size(Sample.results, 2)
    ```

    Will return query object like

    ```python
    {"results": {"$size": 2}}
    ```

    MongoDB doc:
    <https://docs.mongodb.com/manual/reference/operator/query/size/>
    """

    def __init__(
        self,
        field,
        num: int,
    ):
        self.field = field
        self.num = num

    @property
    def query(self):
        return {self.field: {"$size": self.num}}



from abc import abstractmethod


from beanie.odm.bulk import BulkWriter, Operation
from beanie.odm.utils.encoder import Encoder
from typing import (
    Callable,
    List,
    Type,
    TYPE_CHECKING,
    Optional,
    Mapping,
    Any,
    Dict,
    Union,
    Generator,
)

from pymongo.client_session import ClientSession
from pymongo.results import UpdateResult, InsertOneResult

from beanie.odm.interfaces.session import SessionMethods
from beanie.odm.interfaces.update import (
    UpdateMethods,
)
from beanie.odm.operators.update import BaseUpdateOperator
from pymongo import UpdateOne as UpdateOnePyMongo
from pymongo import UpdateMany as UpdateManyPyMongo

if TYPE_CHECKING:
    from beanie.odm.documents import DocType


class UpdateQuery(UpdateMethods, SessionMethods):
    """
    Update Query base class

    Inherited from:

    - [SessionMethods](https://roman-right.github.io/beanie/api/interfaces/#sessionmethods)
    - [UpdateMethods](https://roman-right.github.io/beanie/api/interfaces/#aggregatemethods)
    """

    def __init__(
        self,
        document_model: Type["DocType"],
        find_query: Mapping[str, Any],
    ):
        self.document_model = document_model
        self.find_query = find_query
        self.update_expressions: List[Mapping[str, Any]] = []
        self.session = None
        self.is_upsert = False
        self.upsert_insert_doc: Optional["DocType"] = None
        self.encoders: Dict[Any, Callable[[Any], Any]] = {}
        self.bulk_writer: Optional[BulkWriter] = None
        self.encoders = (
            self.document_model.get_settings().model_settings.bson_encoders
        )
        self.pymongo_kwargs: Dict[str, Any] = {}

    @property
    def update_query(self) -> Dict[str, Any]:
        query: Dict[str, Any] = {}
        for expression in self.update_expressions:
            if isinstance(expression, BaseUpdateOperator):
                query.update(expression.query)
            elif isinstance(expression, dict):
                query.update(expression)
            else:
                raise TypeError("Wrong expression type")
        return Encoder(custom_encoders=self.encoders).encode(query)

    def update(
        self,
        *args: Mapping[str, Any],
        session: Optional[ClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        **pymongo_kwargs,
    ) -> "UpdateQuery":
        """
        Provide modifications to the update query.

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: Optional[ClientSession]
        :param bulk_writer: Optional[BulkWriter]
        :param **pymongo_kwargs: pymongo native parameters for update operation
        :return: UpdateMany query
        """
        self.set_session(session=session)
        self.update_expressions += args
        if bulk_writer:
            self.bulk_writer = bulk_writer
        self.pymongo_kwargs.update(pymongo_kwargs)
        return self

    def upsert(
        self,
        *args: Mapping[str, Any],
        on_insert: "DocType",
        session: Optional[ClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        **pymongo_kwargs,
    ) -> "UpdateQuery":
        """
        Provide modifications to the upsert query.

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param on_insert: DocType - document to insert if there is no matched
        document in the collection
        :param session: Optional[ClientSession]
        :param **pymongo_kwargs: pymongo native parameters for update operation
        :return: UpdateMany query
        """
        self.upsert_insert_doc = on_insert  # type: ignore
        self.update(
            *args, session=session, bulk_writer=bulk_writer, **pymongo_kwargs
        )
        return self

    @abstractmethod
    async def _update(self) -> UpdateResult:
        ...

    def __await__(
        self,
    ) -> Generator[Any, None, Union[UpdateResult, InsertOneResult]]:
        """
        Run the query
        :return:
        """

        update_result = yield from self._update().__await__()
        if self.upsert_insert_doc is None:
            return update_result
        else:
            if update_result is not None and update_result.matched_count == 0:
                return (
                    yield from self.document_model.insert_one(
                        document=self.upsert_insert_doc,
                        session=self.session,
                        bulk_writer=self.bulk_writer,
                    ).__await__()
                )
            else:
                return update_result


class UpdateMany(UpdateQuery):
    """
    Update Many query class

    Inherited from:

    - [UpdateQuery](https://roman-right.github.io/beanie/api/queries/#updatequery)
    """

    def update_many(
        self,
        *args: Mapping[str, Any],
        session: Optional[ClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        **pymongo_kwargs,
    ):
        """
        Provide modifications to the update query

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: Optional[ClientSession]
        :param bulk_writer: "BulkWriter" - Beanie bulk writer
        :param **pymongo_kwargs: pymongo native parameters for update operation
        :return: UpdateMany query
        """
        return self.update(
            *args, session=session, bulk_writer=bulk_writer, **pymongo_kwargs
        )

    async def _update(self):
        if self.bulk_writer is None:
            return (
                await self.document_model.get_motor_collection().update_many(
                    self.find_query,
                    self.update_query,
                    session=self.session,
                    **self.pymongo_kwargs,
                )
            )
        else:
            self.bulk_writer.add_operation(
                Operation(
                    operation=UpdateManyPyMongo,
                    first_query=self.find_query,
                    second_query=self.update_query,
                    object_class=self.document_model,
                )
            )


class UpdateOne(UpdateQuery):
    """
    Update One query class

    Inherited from:

    - [UpdateQuery](https://roman-right.github.io/beanie/api/queries/#updatequery)
    """

    def update_one(
        self,
        *args: Mapping[str, Any],
        session: Optional[ClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        **pymongo_kwargs,
    ):
        """
        Provide modifications to the update query. The same as `update()`

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: Optional[ClientSession]
        :param bulk_writer: "BulkWriter" - Beanie bulk writer
        :param **pymongo_kwargs: pymongo native parameters for update operation
        :return: UpdateMany query
        """
        return self.update(
            *args, session=session, bulk_writer=bulk_writer, **pymongo_kwargs
        )

    async def _update(self):
        if not self.bulk_writer:
            return await self.document_model.get_motor_collection().update_one(
                self.find_query,
                self.update_query,
                session=self.session,
                **self.pymongo_kwargs,
            )
        else:
            self.bulk_writer.add_operation(
                Operation(
                    operation=UpdateOnePyMongo,
                    first_query=self.find_query,
                    second_query=self.update_query,
                    object_class=self.document_model,
                )
            )



from abc import abstractmethod
from typing import Any, Optional, Union, List, Dict, cast

from pymongo.client_session import ClientSession

from beanie.odm.fields import ExpressionField


class AggregateMethods:
    """
    Aggregate methods
    """

    @abstractmethod
    def aggregate(
        self,
        aggregation_pipeline,
        projection_model=None,
        session=None,
        ignore_cache: bool = False,
    ):
        ...

    async def sum(
        self,
        field: Union[str, ExpressionField],
        session: Optional[ClientSession] = None,
        ignore_cache: bool = False,
    ) -> Optional[float]:
        """
        Sum of values of the given field

        Example:

        ```python

        class Sample(Document):
            price: int
            count: int

        sum_count = await Document.find(Sample.price <= 100).sum(Sample.count)

        ```

        :param field: Union[str, ExpressionField]
        :param session: Optional[ClientSession] - pymongo session
        :param ignore_cache: bool
        :return: float - sum. None if there are no items.
        """
        pipeline = [
            {"$group": {"_id": None, "sum": {"$sum": f"${field}"}}},
            {"$project": {"_id": 0, "sum": 1}},
        ]

        # As we did not supply a projection we can safely cast the type (hinting to mypy that we know the type)
        result: List[Dict[str, Any]] = cast(
            List[Dict[str, Any]],
            await self.aggregate(
                aggregation_pipeline=pipeline,
                session=session,
                ignore_cache=ignore_cache,
            ).to_list(),  # type: ignore # TODO: pyright issue, fix
        )
        if not result:
            return None
        return result[0]["sum"]

    async def avg(
        self,
        field,
        session: Optional[ClientSession] = None,
        ignore_cache: bool = False,
    ) -> Optional[float]:
        """
        Average of values of the given field

        Example:

        ```python

        class Sample(Document):
            price: int
            count: int

        avg_count = await Document.find(Sample.price <= 100).avg(Sample.count)
        ```

        :param field: Union[str, ExpressionField]
        :param session: Optional[ClientSession] - pymongo session
        :param ignore_cache: bool
        :return: Optional[float] - avg. None if there are no items.
        """
        pipeline = [
            {"$group": {"_id": None, "avg": {"$avg": f"${field}"}}},
            {"$project": {"_id": 0, "avg": 1}},
        ]

        result: List[Dict[str, Any]] = cast(
            List[Dict[str, Any]],
            await self.aggregate(
                aggregation_pipeline=pipeline,
                session=session,
                ignore_cache=ignore_cache,
            ).to_list(),  # type: ignore # TODO: pyright issue, fix
        )
        if not result:
            return None
        return result[0]["avg"]

    async def max(
        self,
        field: Union[str, ExpressionField],
        session: Optional[ClientSession] = None,
        ignore_cache: bool = False,
    ) -> Optional[float]:
        """
        Max of the values of the given field

        Example:

        ```python

        class Sample(Document):
            price: int
            count: int

        max_count = await Document.find(Sample.price <= 100).max(Sample.count)
        ```

        :param field: Union[str, ExpressionField]
        :param session: Optional[ClientSession] - pymongo session
        :return: float - max. None if there are no items.
        """
        pipeline = [
            {"$group": {"_id": None, "max": {"$max": f"${field}"}}},
            {"$project": {"_id": 0, "max": 1}},
        ]

        result: List[Dict[str, Any]] = cast(
            List[Dict[str, Any]],
            await self.aggregate(
                aggregation_pipeline=pipeline,
                session=session,
                ignore_cache=ignore_cache,
            ).to_list(),  # type: ignore # TODO: pyright issue, fix
        )
        if not result:
            return None
        return result[0]["max"]

    async def min(
        self,
        field: Union[str, ExpressionField],
        session: Optional[ClientSession] = None,
        ignore_cache: bool = False,
    ) -> Optional[float]:
        """
        Min of the values of the given field

        Example:

        ```python

        class Sample(Document):
            price: int
            count: int

        min_count = await Document.find(Sample.price <= 100).min(Sample.count)
        ```

        :param field: Union[str, ExpressionField]
        :param session: Optional[ClientSession] - pymongo session
        :return: float - min. None if there are no items.
        """
        pipeline = [
            {"$group": {"_id": None, "min": {"$min": f"${field}"}}},
            {"$project": {"_id": 0, "min": 1}},
        ]

        result: List[Dict[str, Any]] = cast(
            List[Dict[str, Any]],
            await self.aggregate(
                aggregation_pipeline=pipeline,
                session=session,
                ignore_cache=ignore_cache,
            ).to_list(),  # type: ignore # TODO: pyright issue, fix
        )
        if not result:
            return None
        return result[0]["min"]



import inspect
from typing import Optional

from pydantic.fields import ModelField
from pydantic.typing import get_origin

from beanie.odm.fields import LinkTypes, LinkInfo, Link


def detect_link(field: ModelField) -> Optional[LinkInfo]:
    """
    It detects link and returns LinkInfo if any found.

    :param field: ModelField
    :return: Optional[LinkInfo]
    """
    if field.type_ == Link:
        if field.allow_none is True:
            return LinkInfo(
                field=field.name,
                model_class=field.sub_fields[0].type_,  # type: ignore
                link_type=LinkTypes.OPTIONAL_DIRECT,
            )
        return LinkInfo(
            field=field.name,
            model_class=field.sub_fields[0].type_,  # type: ignore
            link_type=LinkTypes.DIRECT,
        )
    if (
        inspect.isclass(get_origin(field.outer_type_))
        and issubclass(get_origin(field.outer_type_), list)  # type: ignore
        and len(field.sub_fields) == 1  # type: ignore
    ):
        internal_field = field.sub_fields[0]  # type: ignore
        if internal_field.type_ == Link:
            if internal_field.allow_none is True:
                return None
            return LinkInfo(
                field=field.name,
                model_class=internal_field.sub_fields[0].type_,  # type: ignore
                link_type=LinkTypes.LIST,
            )
    return None



from typing import Any, Type, Union, TYPE_CHECKING

from pydantic import BaseModel

if TYPE_CHECKING:
    from beanie.odm.documents import Document


def parse_obj(
    model: Union[Type[BaseModel], Type["Document"]], data: Any
) -> BaseModel:
    if hasattr(model, "_parse_obj_saving_state"):
        return model._parse_obj_saving_state(data)  # type: ignore
    return model.parse_obj(data)



from datetime import timedelta
from typing import Any, Dict, Optional, Type

from pydantic import BaseModel, Field


class ModelSettings(BaseModel):
    projection: Optional[Dict[str, Any]] = None
    use_state_management: bool = False
    state_management_replace_objects: bool = False
    validate_on_save: bool = False
    use_revision: bool = False
    use_cache: bool = False
    cache_capacity: int = 32
    cache_expiration_time: timedelta = timedelta(minutes=10)
    bson_encoders: Dict[Any, Any] = Field(default_factory=dict)

    @classmethod
    def init(
        cls,
        document_model: Type,
    ) -> "ModelSettings":
        settings_class = getattr(document_model, "Settings", None)
        if settings_class is not None:
            return cls.parse_obj(vars(settings_class))
        else:
            return ModelSettings()



from enum import Enum

import pymongo


class SortDirection(int, Enum):
    """
    Sorting directions
    """

    ASCENDING = pymongo.ASCENDING
    DESCENDING = pymongo.DESCENDING


class InspectionStatuses(str, Enum):
    """
    Statuses of the collection inspection
    """

    FAIL = "FAIL"
    OK = "OK"



from typing import List, Type, Optional

from motor.motor_asyncio import AsyncIOMotorCollection, AsyncIOMotorDatabase
from pydantic.main import BaseModel
from pymongo import IndexModel

from beanie.exceptions import MongoDBVersionError
from beanie.odm.settings.timeseries import TimeSeriesConfig


class IndexModelField(IndexModel):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, IndexModel):
            return v
        else:
            return IndexModel(v)


class CollectionInputParameters(BaseModel):
    name: str = ""
    indexes: List[IndexModelField] = []
    timeseries: Optional[TimeSeriesConfig]

    class Config:
        arbitrary_types_allowed = True


class CollectionSettings(BaseModel):
    name: str
    motor_collection: AsyncIOMotorCollection

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    async def init(
        cls,
        database: AsyncIOMotorDatabase,
        document_model: Type,
        allow_index_dropping: bool,
    ) -> "CollectionSettings":
        """
        Collection settings factory.
        Creates private property _collection_settings
        for the Document on the init step,

        :param database: AsyncIOMotorDatabase - Motor database instance
        :param document_model: Type - a class, inherited from Document class
        :param allow_index_dropping: bool - if index dropping is allowed
        :return: CollectionSettings
        """
        # parse collection parameters
        collection_class = getattr(document_model, "Collection", None)
        if collection_class:
            collection_parameters = CollectionInputParameters.parse_obj(
                vars(collection_class)
            )
        else:
            collection_parameters = CollectionInputParameters()

        # set collection name
        if not collection_parameters.name:
            collection_parameters.name = document_model.__name__

        # check mongodb version
        build_info = await database.command({"buildInfo": 1})
        mongo_version = build_info["version"]
        major_version = int(mongo_version.split(".")[0])

        if collection_parameters.timeseries is not None and major_version < 5:
            raise MongoDBVersionError(
                "Timeseries are supported by MongoDB version 5 and higher"
            )

        # create motor collection
        if (
            collection_parameters.timeseries is not None
            and collection_parameters.name
            not in await database.list_collection_names()
        ):

            collection = await database.create_collection(
                **collection_parameters.timeseries.build_query(
                    collection_parameters.name
                )
            )
        else:
            collection = database[collection_parameters.name]

        # indexes
        old_indexes = (await collection.index_information()).keys()
        new_indexes = ["_id_"]

        # Indexed field wrapped with Indexed()
        found_indexes = [
            IndexModel(
                [
                    (
                        fvalue.alias,
                        fvalue.type_._indexed[0],
                    )
                ],
                **fvalue.type_._indexed[1]
            )
            for _, fvalue in document_model.__fields__.items()
            if hasattr(fvalue.type_, "_indexed") and fvalue.type_._indexed
        ]

        # get indexes from the Collection class
        if collection_parameters.indexes:
            found_indexes += collection_parameters.indexes

        # create indices
        if found_indexes:
            new_indexes += await collection.create_indexes(found_indexes)

        # delete indexes
        # Only drop indexes if the user specifically allows for it
        if allow_index_dropping:
            for index in set(old_indexes) - set(new_indexes):
                await collection.drop_index(index)

        return cls(
            name=collection_parameters.name,
            motor_collection=collection,
        )



from abc import abstractmethod
from typing import Any, Mapping

from beanie.odm.operators import BaseOperator


class BaseUpdateOperator(BaseOperator):
    @property
    @abstractmethod
    def query(self) -> Mapping[str, Any]:
        ...



from dataclasses import dataclass
from typing import Type
from motor.motor_asyncio import AsyncIOMotorDatabase

from beanie.odm.settings.collection import CollectionSettings
from beanie.odm.settings.model import ModelSettings


@dataclass
class DocumentSettings:
    model_settings: ModelSettings
    collection_settings: CollectionSettings

    @classmethod
    async def init(cls, database: AsyncIOMotorDatabase, document_model: Type, allow_index_dropping: bool):
        # Init collection settings
        collection_settings = await CollectionSettings.init(
            database=database,
            document_model=document_model,
            allow_index_dropping=allow_index_dropping,
        )

        # Init model settings

        model_settings = ModelSettings.init(document_model=document_model)
        return cls(
            model_settings=model_settings,
            collection_settings=collection_settings,
        )



from beanie.odm.fields import LinkTypes
from typing import TYPE_CHECKING, List, Dict, Any, Type

if TYPE_CHECKING:
    from beanie import Document


def construct_lookup_queries(cls: Type["Document"]) -> List[Dict[str, Any]]:
    queries = []
    link_fields = cls.get_link_fields()
    if link_fields is not None:
        for link_info in link_fields.values():
            if link_info.link_type in [
                LinkTypes.DIRECT,
                LinkTypes.OPTIONAL_DIRECT,
            ]:
                queries += [
                    {
                        "$lookup": {
                            "from": link_info.model_class.get_motor_collection().name,  # type: ignore
                            "localField": f"{link_info.field}.$id",
                            "foreignField": "_id",
                            "as": f"_link_{link_info.field}",
                        }
                    },
                    {
                        "$unwind": {
                            "path": f"$_link_{link_info.field}",
                            "preserveNullAndEmptyArrays": True,
                        }
                    },
                    {
                        "$set": {
                            link_info.field: {
                                "$cond": {
                                    "if": {
                                        "$ifNull": [
                                            f"$_link_{link_info.field}",
                                            False,
                                        ]
                                    },
                                    "then": f"$_link_{link_info.field}",
                                    "else": f"${link_info.field}",
                                }
                            }
                        }
                    },
                ]  # type: ignore
            else:
                queries.append(
                    {
                        "$lookup": {
                            "from": link_info.model_class.get_motor_collection().name,  # type: ignore
                            "localField": f"{link_info.field}.$id",
                            "foreignField": "_id",
                            "as": link_info.field,
                        }
                    }
                )
    return queries



from abc import ABC
from enum import Enum
from typing import List, Optional

from beanie.odm.operators.find import BaseFindOperator


class BaseFindGeospatialOperator(BaseFindOperator, ABC):
    ...


class GeoIntersects(BaseFindGeospatialOperator):
    """
    `$geoIntersects` query operator

    Example:

    ```python
    class GeoObject(BaseModel):
        type: str = "Point"
        coordinates: Tuple[float, float]

    class Place(Document):
        geo: GeoObject

        class Collection:
            name = "places"
            indexes = [
                [("geo", pymongo.GEOSPHERE)],  # GEO index
            ]

    GeoIntersects(Place.geo, "Polygon", [[0,0], [1,1], [3,3]])
    ```

    Will return query object like

    ```python
    {
        "geo": {
            "$geoIntersects": {
                "$geometry": {
                    "type": "Polygon",
                    "coordinates": [[0,0], [1,1], [3,3]],
                }
            }
        }
    }
    ```

    MongoDB doc:
    <https://docs.mongodb.com/manual/reference/operator/query/geoIntersects/>
    """

    def __init__(self, field, geo_type: str, coordinates: List[List[float]]):
        self.field = field
        self.geo_type = geo_type
        self.coordinates = coordinates

    @property
    def query(self):
        return {
            self.field: {
                "$geoIntersects": {
                    "$geometry": {
                        "type": self.geo_type,
                        "coordinates": self.coordinates,
                    }
                }
            }
        }


class GeoWithinTypes(str, Enum):
    Polygon = "Polygon"
    MultiPolygon = "MultiPolygon"


class GeoWithin(BaseFindGeospatialOperator):
    """
    `$geoWithin` query operator

    Example:

    ```python
    class GeoObject(BaseModel):
        type: str = "Point"
        coordinates: Tuple[float, float]

    class Place(Document):
        geo: GeoObject

        class Collection:
            name = "places"
            indexes = [
                [("geo", pymongo.GEOSPHERE)],  # GEO index
            ]

    GeoWithin(Place.geo, "Polygon", [[0,0], [1,1], [3,3]])
    ```

    Will return query object like

    ```python
    {
        "geo": {
            "$geoWithin": {
                "$geometry": {
                    "type": "Polygon",
                    "coordinates": [[0,0], [1,1], [3,3]],
                }
            }
        }
    }
    ```

    MongoDB doc:
    <https://docs.mongodb.com/manual/reference/operator/query/geoWithin/>
    """

    def __init__(
        self, field, geo_type: GeoWithinTypes, coordinates: List[List[float]]
    ):
        self.field = field
        self.geo_type = geo_type
        self.coordinates = coordinates

    @property
    def query(self):
        return {
            self.field: {
                "$geoWithin": {
                    "$geometry": {
                        "type": self.geo_type,
                        "coordinates": self.coordinates,
                    }
                }
            }
        }


class Near(BaseFindGeospatialOperator):
    """
    `$near` query operator

    Example:

    ```python
    class GeoObject(BaseModel):
        type: str = "Point"
        coordinates: Tuple[float, float]

    class Place(Document):
        geo: GeoObject

        class Collection:
            name = "places"
            indexes = [
                [("geo", pymongo.GEOSPHERE)],  # GEO index
            ]

    Near(Place.geo, 1.2345, 2.3456, min_distance=500)
    ```

    Will return query object like

    ```python
    {
        "geo": {
            "$near": {
                "$geometry": {
                    "type": "Point",
                    "coordinates": [1.2345, 2.3456],
                },
                "$maxDistance": 500,
            }
        }
    }
    ```

    MongoDB doc:
    <https://docs.mongodb.com/manual/reference/operator/query/near/>
    """

    operator = "$near"

    def __init__(
        self,
        field,
        longitude: float,
        latitude: float,
        max_distance: Optional[float] = None,
        min_distance: Optional[float] = None,
    ):
        self.field = field
        self.longitude = longitude
        self.latitude = latitude
        self.max_distance = max_distance
        self.min_distance = min_distance

    @property
    def query(self):
        expression = {
            self.field: {
                self.operator: {
                    "$geometry": {
                        "type": "Point",
                        "coordinates": [self.longitude, self.latitude],
                    },
                }
            }
        }
        if self.max_distance:
            expression[self.field][self.operator][
                "$maxDistance"
            ] = self.max_distance  # type: ignore
        if self.min_distance:
            expression[self.field][self.operator][
                "$minDistance"
            ] = self.min_distance  # type: ignore
        return expression


class NearSphere(Near):
    """
    `$nearSphere` query operator

    Example:

    ```python
    class GeoObject(BaseModel):
        type: str = "Point"
        coordinates: Tuple[float, float]

    class Place(Document):
        geo: GeoObject

        class Collection:
            name = "places"
            indexes = [
                [("geo", pymongo.GEOSPHERE)],  # GEO index
            ]

    NearSphere(Place.geo, 1.2345, 2.3456, min_distance=500)
    ```

    Will return query object like

    ```python
    {
        "geo": {
            "$nearSphere": {
                "$geometry": {
                    "type": "Point",
                    "coordinates": [1.2345, 2.3456],
                },
                "$maxDistance": 500,
            }
        }
    }
    ```

    MongoDB doc:
    <https://docs.mongodb.com/manual/reference/operator/query/nearSphere/>
    """

    operator = "$nearSphere"



from collections import deque
from datetime import datetime, timedelta
from decimal import Decimal
from enum import Enum
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from pathlib import PurePath
from types import GeneratorType
from typing import (
    AbstractSet,
    List,
    Mapping,
    Union,
)
from typing import Any, Callable, Dict, Type
from uuid import UUID

from bson import ObjectId, DBRef, Binary
from pydantic import BaseModel
from pydantic import SecretBytes, SecretStr
from pydantic.color import Color

from beanie.odm.fields import Link, LinkTypes
from beanie.odm import documents

ENCODERS_BY_TYPE: Dict[Type[Any], Callable[[Any], Any]] = {
    Color: str,
    timedelta: lambda td: td.total_seconds(),
    Decimal: float,
    deque: list,
    IPv4Address: str,
    IPv4Interface: str,
    IPv4Network: str,
    IPv6Address: str,
    IPv6Interface: str,
    IPv6Network: str,
    SecretBytes: SecretBytes.get_secret_value,
    SecretStr: SecretStr.get_secret_value,
    Enum: lambda o: o.value,
    PurePath: str,
    Link: lambda l: l.ref,
    bytes: lambda b: b if isinstance(b, Binary) else Binary(b),
}


class Encoder:
    """
    BSON encoding class
    """

    def __init__(
        self,
        exclude: Union[
            AbstractSet[Union[str, int]], Mapping[Union[str, int], Any], None
        ] = None,
        custom_encoders: Dict[Type, Callable] = None,
        by_alias: bool = True,
        to_db: bool = False,
    ):
        self.exclude = exclude or {}
        self.by_alias = by_alias
        self.custom_encoders = custom_encoders or {}
        self.to_db = to_db

    def encode(self, obj: Any):
        """
        Run the encoder
        """
        return self._encode(obj=obj)

    def encode_document(self, obj):
        """
        Beanie Document class case
        """
        encoder = Encoder(
            exclude=self.exclude,
            custom_encoders=obj.get_settings().model_settings.bson_encoders,
            by_alias=self.by_alias,
            to_db=self.to_db,
        )

        link_fields = obj.get_link_fields()
        obj_dict: Dict[str, Any] = {}
        for k, o in obj._iter(to_dict=False, by_alias=self.by_alias):
            if k not in self.exclude:
                if link_fields and k in link_fields:
                    if link_fields[k].link_type == LinkTypes.LIST:
                        obj_dict[k] = [link.to_ref() for link in o]
                    if link_fields[k].link_type == LinkTypes.DIRECT:
                        obj_dict[k] = o.to_ref()
                    if link_fields[k].link_type == LinkTypes.OPTIONAL_DIRECT:
                        if o is not None:
                            obj_dict[k] = o.to_ref()
                        else:
                            obj_dict[k] = o
                else:
                    obj_dict[k] = o
                obj_dict[k] = encoder.encode(obj_dict[k])
        return obj_dict

    def encode_base_model(self, obj):
        """
        BaseModel case
        """
        obj_dict = {}
        for k, o in obj._iter(to_dict=False, by_alias=self.by_alias):
            if k not in self.exclude:
                obj_dict[k] = self._encode(o)

        return obj_dict

    def encode_dict(self, obj):
        """
        Dictionary case
        """
        for key, value in obj.items():
            obj[key] = self._encode(value)
        return obj

    def encode_iterable(self, obj):
        """
        Iterable case
        """
        return [self._encode(item) for item in obj]

    def _encode(
        self,
        obj,
    ) -> Any:
        """"""

        if self.custom_encoders:
            if type(obj) in self.custom_encoders:
                return self.custom_encoders[type(obj)](obj)
            for encoder_type, encoder in self.custom_encoders.items():
                if isinstance(obj, encoder_type):
                    return encoder(obj)
        if type(obj) in ENCODERS_BY_TYPE:
            return ENCODERS_BY_TYPE[type(obj)](obj)
        for cls, encoder in ENCODERS_BY_TYPE.items():
            if isinstance(obj, cls):
                return encoder(obj)

        if isinstance(obj, documents.Document):
            return self.encode_document(obj)
        if isinstance(obj, BaseModel):
            return self.encode_base_model(obj)
        if isinstance(obj, dict):
            return self.encode_dict(obj)
        if isinstance(obj, (list, set, frozenset, GeneratorType, tuple)):
            return self.encode_iterable(obj)

        if isinstance(
            obj, (str, int, float, ObjectId, UUID, datetime, type(None), DBRef)
        ):
            return obj

        errors: List[Exception] = []
        try:
            data = dict(obj)
        except Exception as e:
            errors.append(e)
            try:
                data = vars(obj)
            except Exception as e:
                errors.append(e)
                raise ValueError(errors)
        return self._encode(data)
